from reflex_clerk_api.base import ClerkBase

Javascript = str
JSX = str
SignInInitialValues = dict[str, str]
SignUpInitialValues = dict[str, str]


class ClerkLoaded(ClerkBase):
    """Only renders children after authentication has been checked."""

    tag = "ClerkLoaded"


class ClerkLoading(ClerkBase):
    """Only renders childen while Clerk authenticates the user."""

    tag = "ClerkLoading"


class Protect(ClerkBase):
    tag = "Protect"

    condition: Javascript | None = None
    "Optional conditional logic that renders the children if it returns true"
    fallback: JSX | None = None
    "An optional snippet of JSX to show when a user doesn't have the role or permission to access the protected content."
    permission: str | None = None
    "Optional string corresponding to a Role's Permission in the format org:<resource>:<action>"
    role: str | None = None
    "Optional string corresponding to an Organization's Role in the format org:<role>"


class RedirectToSignIn(ClerkBase):
    """Immediately redirects the user to the sign in page when rendered."""

    tag = "RedirectToSignIn"

    signInFallbackRedirectUrl: str | None = None
    "The fallback URL to redirect to after the user signs in, if there's no redirect_url in the path already. Defaults to /."
    signInForceRedirectUrl: str | None = None
    "If provided, this URL will always be redirected to after the user signs in."
    initialValues: SignInInitialValues | None = None
    "The values used to prefill the sign-in fields with."


class RedirectToSignUp(ClerkBase):
    """Immediately redirects the user to the sign up page when rendered."""

    tag = "RedirectToSignUp"

    signUpFallbackRedirectUrl: str | None = None
    "The fallback URL to redirect to after the user signs up, if there's no redirect_url in the path already. Defaults to /."
    signUpForceRedirectUrl: str | None = None
    "If provided, this URL will always be redirected to after the user signs up."
    initialValues: SignUpInitialValues | None = None
    "The values used to prefill the sign-up fields with."


class RedirectToUserProfile(ClerkBase):
    """Immediately redirects the user to their profile page when rendered."""

    tag = "RedirectToUserProfile"


class RedirectToOrganizationProfile(ClerkBase):
    tag = "RedirectToOrganizationProfile"


class RedirectToCreateOrganization(ClerkBase):
    tag = "RedirectToCreateOrganization"


class SignedIn(ClerkBase):
    """Only renders children when the user is signed in."""

    tag = "SignedIn"


class SignedOut(ClerkBase):
    """Only renders children when the user is signed out."""

    tag = "SignedOut"


clerk_loaded = ClerkLoaded.create
clerk_loading = ClerkLoading.create
protect = Protect.create
redirect_to_sign_in = RedirectToSignIn.create
redirect_to_sign_up = RedirectToSignUp.create
redirect_to_user_profile = RedirectToUserProfile.create
redirect_to_organization_profile = RedirectToOrganizationProfile.create
redirect_to_create_organization = RedirectToCreateOrganization.create
signed_in = SignedIn.create
signed_out = SignedOut.create



import logging
import os
import uuid
from pathlib import Path
from typing import Iterator

import clerk_backend_api
import pytest
from clerk_backend_api import Clerk, User
from dotenv import load_dotenv
from playwright.sync_api import Page, expect
from reflex.testing import AppHarness

load_dotenv()

TEST_EMAIL = "ci-test+clerk_test@gmail.com"
TEST_PASSWORD = "test-clerk-password"

PAGE_LOAD_TIMEOUT = 10000 if not os.getenv("CI") else 30000
INTERACTION_TIMEOUT = 2000 if not os.getenv("CI") else 10000


@pytest.fixture(scope="session")
def demo_app():
    app_root = Path(__file__).parent.parent / "clerk_api_demo"
    with AppHarness.create(root=app_root) as harness:
        yield harness


@pytest.fixture(scope="function")
def page(
    request: pytest.FixtureRequest, demo_app: AppHarness, page: Page
) -> Iterator[Page]:
    """Load the demo app main page."""
    page.set_default_timeout(PAGE_LOAD_TIMEOUT)
    assert demo_app.frontend_url is not None
    page.goto(demo_app.frontend_url)
    page.set_default_timeout(INTERACTION_TIMEOUT)
    yield page
    if request.session.testsfailed:
        logging.error("Test failed. Saving screenshot as playwright_test_error.png")
        page.screenshot(path="playwright_test_error.png")


@pytest.fixture()
def clerk_client() -> Iterator[Clerk]:
    """Create clerk backend api client."""
    secret_key = os.environ["CLERK_SECRET_KEY"]
    with clerk_backend_api.Clerk(bearer_auth=secret_key) as client:
        yield client


@pytest.fixture
def create_test_user(clerk_client: Clerk) -> User:
    """Creates (or checks already exists) a test clerk user.

    This can then be used to sign in during tests.
    """
    existing = clerk_client.users.list(request={"email_address": [TEST_EMAIL]})
    if existing is not None and len(existing) > 0:
        if len(existing) != 1:
            raise SetupError(
                "Multiple test users found with same email... This should not happen."
            )
        logging.error("Test user already exists.")
        return existing[0]

    logging.error("Creating test user...")
    res = clerk_client.users.create(
        request={
            "external_id": "ext-id-" + uuid.uuid4().hex[:5],
            "first_name": "John",
            "last_name": "Doe",
            "email_address": [
                TEST_EMAIL,
            ],
            "username": "fake_username_" + uuid.uuid4().hex[:5],
            "password": TEST_PASSWORD,
            "skip_password_checks": False,
            "skip_password_requirement": False,
            "public_metadata": {
                "role": "user",
            },
            "private_metadata": {
                "internal_id": "789",
            },
            "unsafe_metadata": {
                "preferences": {
                    "theme": "dark",
                },
            },
            "delete_self_enabled": True,
            "skip_legal_checks": False,
            "create_organization_enabled": True,
            "create_organizations_limit": 134365,
            "created_at": "2023-03-15T07:15:20.902Z",
        }
    )
    assert res is not None
    return res


def test_test_user(create_test_user: User):
    """Check the test user was either created or found correctly."""
    user = create_test_user
    assert user.email_addresses is not None
    assert user.email_addresses[0].email_address == TEST_EMAIL


def test_render(page: Page):
    """Check that the demo renders correctly.

    I.e. Check components are visible.
    """
    page.pause()
    expect(page.locator('[id="__next"]')).to_contain_text("reflex-clerk-api demo")
    expect(page.locator('[id="__next"]')).to_contain_text("Getting Started")
    expect(page.locator('[id="__next"]')).to_contain_text("Demos")

    expect(page.get_by_test_id("clerkstate_variables_and_methods")).to_be_visible()
    expect(page.get_by_test_id("clerk_loaded_and_signed_in_out_areas")).to_be_visible()
    expect(page.get_by_test_id("better_on_load_handling")).to_be_visible()
    expect(page.get_by_test_id("on_auth_change_callbacks")).to_be_visible()
    expect(page.get_by_test_id("clerkuser_info")).to_be_visible()
    expect(page.get_by_test_id("sign-in_and_sign-up_pages")).to_be_visible()
    expect(page.get_by_test_id("user_profile_management")).to_be_visible()


def test_sign_up(page: Page):
    """Check sign-up button takes you to a sign-up form.

    Note: Can't actually test signing up in headless mode because of bot detection.
    """
    page.pause()
    page.get_by_role("button", name="Sign up").click()
    expect(page.get_by_role("heading")).to_contain_text("Create your account")


class SetupError(Exception):
    """Error raised when the test setup is incorrect."""


@pytest.mark.usefixtures("create_test_user")
def test_sign_in(page: Page):
    """Check sign-in button takes you to a sign-in form.

    Note: Can't actually test signing in in headless mode because of bot detection.
    """

    page.get_by_role("button", name="Sign in").click()
    expect(page.get_by_role("heading")).to_contain_text("Sign in to")
    page.get_by_role("textbox", name="Email address").click()
    page.get_by_role("textbox", name="Email address").fill(TEST_EMAIL)
    page.pause()
    page.get_by_role("button", name="Continue", exact=True).click()
    page.get_by_role("textbox", name="Password").click()
    page.get_by_role("textbox", name="Password").fill(TEST_PASSWORD)
    page.get_by_role("button", name="Continue", exact=True).click()
    expect(page.get_by_test_id("sign_out")).not_to_be_visible()

    page.pause()


@pytest.fixture
def sign_in(page: Page, create_test_user: User) -> User:
    """Sign in the test user."""
    assert create_test_user.email_addresses is not None
    assert create_test_user.email_addresses[0].email_address == TEST_EMAIL
    page.get_by_role("button", name="Sign in").click()
    page.get_by_role("textbox", name="Email address").click()
    page.get_by_role("textbox", name="Email address").fill(TEST_EMAIL)
    page.get_by_role("button", name="Continue", exact=True).click()
    page.get_by_role("textbox", name="Password").click()
    page.get_by_role("textbox", name="Password").fill(TEST_PASSWORD)
    page.get_by_role("button", name="Continue", exact=True).click()
    # Wait until we are back at the demo page signed in
    expect(page.get_by_test_id("sign_out")).not_to_be_visible()

    return create_test_user


@pytest.mark.usefixtures("sign_in")
def test_sign_out(page: Page):
    """Check sign-out button signs out the user."""
    page.get_by_role("button", name="Sign out").click()
    expect(page.get_by_test_id("sign_in")).to_be_visible()
    expect(page.get_by_test_id("sign_up")).to_be_visible()
    expect(page.get_by_test_id("sign_out")).not_to_be_visible()


def test_signed_in_state(page: Page, sign_in: User):
    """Check a signed-in user sees expected state of app."""
    assert sign_in.id is not None
    page.get_by_test_id("clerkstate_variables_and_methods").hover()
    page.pause()
    expect(page.get_by_test_id("is_hydrated")).to_contain_text("true")
    expect(page.get_by_test_id("auth_checked")).to_contain_text("true")
    expect(page.get_by_test_id("is_signed_in")).to_contain_text("true")
    expect(page.get_by_test_id("user_id")).to_contain_text(sign_in.id)

    page.get_by_test_id("clerk_loaded_and_signed_in_out_areas").hover()
    page.pause()
    expect(page.get_by_test_id("you_are_signed_in")).to_contain_text(
        "You are signed in."
    )
    expect(page.get_by_test_id("you_are_signed_in")).to_be_visible()
    expect(page.get_by_test_id("you_are_signed_out")).not_to_be_visible()

    page.get_by_test_id("better_on_load_handling").hover()
    page.pause()
    expect(page.get_by_test_id("info_from_load")).to_contain_text(
        "clerkstate.auth_checked: True"
    )



"""
This provides some example code for adding dedicated sign-in and sign-up pages to your app.
"""

import os

import reflex as rx

import reflex_clerk_api as clerk


def add_sign_in_page(
    app: rx.App, publishable_key: str | None = None, route: str = "/sign-in"
) -> None:
    """
    Adds a sign-in page that is customizable via the Clerk dashboard.
    """
    assert route.startswith("/")
    publishable_key = publishable_key or os.environ["CLERK_PUBLISHABLE_KEY"]

    sign_in_page = clerk.clerk_provider(
        rx.center(
            rx.vstack(
                clerk.sign_in(path=route),
                align="center",
                spacing="7",
            ),
            height="100vh",
        ),
        publishable_key=publishable_key,
    )
    app.add_page(sign_in_page, route=route + "/[[...signin]]")


def add_sign_up_page(
    app: rx.App, publishable_key: str | None = None, route: str = "/sign-up"
) -> None:
    """
    Adds a sign-up page that is customizable via the Clerk dashboard.
    """
    assert route.startswith("/")
    publishable_key = publishable_key or os.environ["CLERK_PUBLISHABLE_KEY"]

    sign_up_page = clerk.clerk_provider(
        rx.center(
            rx.vstack(
                clerk.sign_up(path=route),
                align="center",
                spacing="7",
            ),
            height="100vh",
        ),
        publishable_key=publishable_key,
    )
    app.add_page(sign_up_page, route=route + "/[[...signup]]")



from reflex_clerk_api.base import ClerkBase


class UserButton(ClerkBase):
    tag = "UserButton"


class UserProfile(ClerkBase):
    tag = "UserProfile"


user_button = UserButton.create
user_profile = UserProfile.create



from reflex_clerk_api.base import ClerkBase


class SignIn(ClerkBase):
    tag = "SignIn"

    path: str


class SignUp(ClerkBase):
    tag = "SignUp"

    path: str


sign_in = SignIn.create
sign_up = SignUp.create



from reflex_clerk_api.base import ClerkBase


class SignUpButton(ClerkBase):
    tag = "SignUpButton"


class SignInButton(ClerkBase):
    tag = "SignInButton"


class SignOutButton(ClerkBase):
    tag = "SignOutButton"


sign_up_button = SignUpButton.create
sign_in_button = SignInButton.create
sign_out_button = SignOutButton.create



import asyncio
import logging
import os
import time
import uuid
from typing import Any, Callable, ClassVar, TypeVar

import authlib.jose.errors as jose_errors
import clerk_backend_api
import reflex as rx
from authlib.jose import JWTClaims, jwt
from reflex.event import EventCallback, EventType, IndividualEventType
from reflex.utils.exceptions import ImmutableStateError

from reflex_clerk_api.base import ClerkBase

from .models import Appearance


class ReflexClerkApiError(Exception):
    pass


class MissingSecretKeyError(ReflexClerkApiError):
    pass


class MissingUserError(ReflexClerkApiError):
    pass


class ClerkState(rx.State):
    is_signed_in: bool = False
    """Whether the user is logged in."""

    auth_checked: bool = False
    """Whether the auth state of the user has been checked yet.
    I.e., has Clerk sent a response to the frontend yet."""

    claims: JWTClaims | None = None
    """The JWT claims of the user, if they are logged in."""
    user_id: str | None = None
    """The clerk user ID of the user, if they are logged in."""

    # NOTE: ClassVar tells reflex it doesn't need to include these in the persisted state per instance.
    _auth_wait_timeout_seconds: ClassVar[float] = 1.0
    _secret_key: ClassVar[str | None] = None
    """The Clerk secret_key set during clerk_provider creation."""
    _on_load_events: ClassVar[dict[uuid.UUID, EventType[()]]] = {}
    _dependent_handlers: ClassVar[dict[int, EventCallback]] = {}
    _client: ClassVar[clerk_backend_api.Clerk | None] = None
    _jwk_keys: ClassVar[dict[str, Any] | None] = None
    _last_jwk_reset: ClassVar[float] = 0.0
    _claims_options: ClassVar[dict[str, Any]] = {
        # "iss": {"value": "https://<your-iss>.clerk.accounts.dev"},
        "exp": {"essential": True},
        "nbf": {"essential": True},
        # "azp": {"essential": False, "values": ["http://localhost:3000", "https://example.com"]},
    }

    @classmethod
    def register_dependent_handler(cls, handler: EventCallback) -> None:
        """Register a handler to be called any time this state updates.

        I.e. Any events that should be triggered on login/logout.
        """
        assert isinstance(handler, rx.EventHandler)
        hash_id = hash((handler.state_full_name, handler.fn))
        logging.debug(f"Dependent hash_id: {hash_id}")
        cls._dependent_handlers[hash_id] = handler

    @classmethod
    def set_auth_wait_timeout_seconds(cls, seconds: float) -> None:
        """Sets the max time to wait for initial auth check before running other on_load events.

        Note: on_load events will still be run after a timed out auth check.
        Check ClerkState.auth_checked to see if auth check is complete.
        """
        cls._auth_wait_timeout_seconds = seconds

    @classmethod
    def set_claims_options(cls, claims_options: dict[str, Any]) -> None:
        """Set the claims options for the JWT claims validation."""
        cls._claims_options = claims_options

    @property
    def client(self) -> clerk_backend_api.Clerk:
        if self._client is None:
            self._set_client()
        assert self._client is not None
        return self._client

    @rx.event
    async def set_clerk_session(self, token: str) -> EventType:
        """Manually obtain user session information via the Clerk JWT.

        This event is triggered by the frontend via the ClerkSessionSynchronizer/ClerkProvider component.
        """
        logging.debug("Setting Clerk session")
        jwks = await self._get_jwk_keys()
        try:
            decoded: JWTClaims = jwt.decode(
                token, {"keys": jwks}, claims_options=self._claims_options
            )
        except jose_errors.DecodeError as e:
            # E.g. DecodeError -- Something went wrong just getting the JWT
            # On next attempt, new JWKs will be fetched
            self.auth_error = e
            self._request_jwk_reset()
            logging.warning(f"JWT decode error: {e}")
            return ClerkState.clear_clerk_session
        try:
            # Validate the token according to the claim options (e.g. iss, exp, nbf, azp.)
            decoded.validate()
        except (jose_errors.InvalidClaimError, jose_errors.MissingClaimError) as e:
            logging.warning(f"JWT token is invalid: {e}")
            return ClerkState.clear_clerk_session

        self.is_signed_in = True
        self.claims = decoded
        self.user_id = str(decoded.get("sub"))
        self.auth_checked = True
        return list(self._dependent_handlers.values())

    @rx.event
    def clear_clerk_session(self) -> EventType:
        """Clear the Clerk session information.

        This event is triggered by the frontend via the ClerkSessionSynchronizer/ClerkProvider component.
        """
        logging.debug("Clearing Clerk session")
        self.reset()
        self.auth_checked = True
        return list(self._dependent_handlers.values())

    @rx.event(background=True)
    async def wait_for_auth_check(self, uid: uuid.UUID | str) -> EventType:
        """Wait for the Clerk authentication to complete (event sent from frontend).

        Can't just use a blocking wait_for_auth_check because we are really waiting for the frontend event trigger to run, so we need to not block that while we wait.

        This can then return on_load events once auth_checked is True.
        """
        uid = uuid.UUID(uid) if isinstance(uid, str) else uid
        logging.debug(f"Waiting for auth check: {uid} ({type(uid)})")

        on_loads = self._on_load_events.get(uid, None)
        if on_loads is None:
            logging.warning("Waited for auth, but no on_load events registered.")
            on_loads = []

        start_time = time.time()
        while time.time() - start_time < self._auth_wait_timeout_seconds:
            if self.auth_checked:
                logging.debug("Auth check complete")
                return on_loads
            logging.debug("...waiting for auth...")
            # TODO: Ideally, wait on some event instead of sleeping
            await asyncio.sleep(0.05)
        logging.warning("Auth check timed out")
        return on_loads

    @classmethod
    def _set_secret_key(cls, secret_key: str) -> None:
        if not secret_key:
            raise MissingSecretKeyError("secret_key must be set (and not empty)")
        cls._secret_key = secret_key

    @classmethod
    def _set_on_load_events(cls, uid: uuid.UUID, on_load_events: EventType[()]) -> None:
        logging.debug(f"Registing on_load events: {uid}")
        cls._on_load_events[uid] = on_load_events

    @classmethod
    def _set_client(cls) -> None:
        if cls._secret_key:
            secret_key = cls._secret_key
        else:
            if "CLERK_SECRET_KEY" not in os.environ:
                raise MissingSecretKeyError(
                    "CLERK_SECRET_KEY either needs to be passed into clerk_provider(...) or set as an environment variable."
                )
            secret_key = os.environ["CLERK_SECRET_KEY"]
        client = clerk_backend_api.Clerk(bearer_auth=secret_key)
        cls._client = client

    @classmethod
    def _set_jwk_keys(cls, keys: dict[str, Any] | None) -> None:
        cls._jwk_keys = keys

    @classmethod
    def _request_jwk_reset(cls) -> None:
        """Reset the JWK keys so they will be re-fetched on next attempt.

        Only do so if it has been a while since last reset (to prevent malicious tokens from forcing
        constant re-fetching).
        """
        now = time.time()
        if now - cls._last_jwk_reset < 10:
            logging.warning("JWK reset requested too soon")
            return
        cls._last_jwk_reset = time.time()
        cls._jwk_keys = None

    async def _get_jwk_keys(self) -> dict[str, Any]:
        """Get the JWK keys from the Clerk API.

        Note: Cannot be a property because it requires async call to populate.
        Only needs to be done once (will be refreshed on errors).
        """
        if self._jwk_keys:
            return self._jwk_keys
        jwks = await self.client.jwks.get_async()
        assert jwks is not None
        assert jwks.keys is not None
        keys = jwks.model_dump()["keys"]
        self._set_jwk_keys(keys)
        return keys

    # @rx.event
    # def force_reset(self) -> None:
    #     """Force a reset of the Clerk state.
    #
    #     E.g. During development testing.
    #     """
    #     self.reset()
    #     self._set_jwk_keys(None)
    #     return rx.toast.success("Forced reset complete.")


class NotRegisteredError(ReflexClerkApiError):
    pass


class ClerkUser(rx.State):
    """Convenience class for using Clerk User information.

    This only contains a subset of the information available. Create your own state if you need more.

    Note: For this to be updated on login/logout events, it must be registered on the ClerkState.
    """

    first_name: str = ""
    last_name: str = ""
    username: str = ""
    email_address: str = ""
    has_image: bool = False
    image_url: str = ""

    # Set to True when the state is registered on the ClerkState to avoid registering it multiple times.
    _is_registered: ClassVar[bool] = False

    @rx.event
    async def load_user(self) -> None:
        try:
            user: clerk_backend_api.models.User = await get_user(self)
        except MissingUserError:
            logging.debug("Clearing user state")
            self.reset()
            return

        logging.debug("Updating user state")
        self.first_name = (
            user.first_name
            if user.first_name and user.first_name != clerk_backend_api.UNSET
            else ""
        )
        self.last_name = (
            user.last_name
            if user.last_name and user.last_name != clerk_backend_api.UNSET
            else ""
        )
        self.username = (
            user.username
            if user.username and user.username != clerk_backend_api.UNSET
            else ""
        )
        self.email_address = (
            user.email_addresses[0].email_address if user.email_addresses else ""
        )
        self.has_image = True if user.has_image is True else False
        self.image_url = user.image_url or ""


class ClerkSessionSynchronizer(rx.Component):
    """ClerkSessionSynchronizer component.

    This is slightly adapted from Elliot Kroo's reflex-clerk.
    """

    tag = "ClerkSessionSynchronizer"

    def add_imports(
        self,
    ) -> rx.ImportDict:
        addl_imports: rx.ImportDict = {
            "@clerk/clerk-react": ["useAuth"],
            "react": ["useContext", "useEffect"],
            "/utils/context": ["EventLoopContext"],
            "/utils/state": ["Event"],
        }
        return addl_imports

    def add_custom_code(self) -> list[str]:
        clerk_state_name = ClerkState.get_full_name()

        return [
            """
function ClerkSessionSynchronizer({ children }) {
  const { getToken, isLoaded, isSignedIn } = useAuth()
  const [ addEvents, connectErrors ] = useContext(EventLoopContext)

  useEffect(() => {
      if (isLoaded && !!addEvents) {
        if (isSignedIn) {
          getToken().then(token => {
            addEvents([Event("%s.set_clerk_session", {token})])
          })
        } else {
          addEvents([Event("%s.clear_clerk_session")])
        }
      }
  }, [isSignedIn])

  return (
      <>{children}</>
  )
}
"""
            % (clerk_state_name, clerk_state_name)
        ]


InitialState = dict[str, Any]
# Should this be EventSpec instead?
JSCallable = Callable


class ClerkProvider(ClerkBase):
    """ClerkProvider component."""

    # The React component tag.
    tag = "ClerkProvider"

    # NOTE: This might be relevant to getting apperance.base_theme to work.
    # lib_dependencies: list[str] = ["@clerk/themes"]
    # def add_imports(self) -> rx.ImportDict:
    #     return {
    #         "@clerk/themes": ["dark", "neobrutalism", "shadesOfPurple"],
    #     }

    # The props of the React component.
    # Note: when Reflex compiles the component to Javascript,
    # `snake_case` property names are automatically formatted as `camelCase`.
    # The prop names may be defined in `camelCase` as well.
    # some_prop: rx.Var[str] = "some default value"
    # some_other_prop: rx.Var[int] = 1

    # Event triggers declaration if any.
    # Below is equivalent to merging `{ "on_change": lambda e: [e] }`
    # onto the default event triggers of parent/base Component.
    # The function defined for the `on_change` trigger maps event for the javascript
    # trigger to what will be passed to the backend event handler function.
    # on_change: rx.EventHandler[lambda e: [e]]

    after_multi_session_single_sign_out_url: str = ""
    """The URL to navigate to after a successful sign-out from multiple sessions."""

    after_sign_out_url: str = ""
    """The full URL or path to navigate to after a successful sign-out."""

    allowed_redirect_origins: list[str | str] = []
    """An optional list of domains to validate user-provided redirect URLs against."""

    allowed_redirect_protocols: list[str] = []
    """An optional list of protocols to validate user-provided redirect URLs against."""

    # NOTE: `apperance.base_theme` does not work yet.
    appearance: Appearance | None = None
    """Optional object to style your components. Will only affect Clerk components."""

    clerk_js_url: str = ""
    """Define the URL that @clerk/clerk-js should be hot-loaded from."""

    clerk_js_variant: str | None = None
    """If your web application only uses control components, set this to 'headless'."""

    clerk_js_version: str = ""
    """Define the npm version for @clerk/clerk-js."""

    # domain: str | JSCallable[[str], bool] = ""
    domain: str = ""
    """Required if your application is a satellite application. Sets the domain."""

    dynamic: bool = False
    """(For Next.js only) Indicates whether Clerk should make dynamic auth data available."""

    # initial_state: InitialState | None = None
    # """Provide an initial state of the Clerk client during server-side rendering."""

    # is_satellite: bool | JSCallable[[str], bool] = False
    is_satellite: bool = False
    """Whether the application is a satellite application."""

    # Not implemented
    # localization: Localization | None = None
    # See https://clerk.com/docs/customization/localization#clerk-localizations for more info.
    # """Optional object to localize your components. Will only affect Clerk components."""

    nonce: str = ""
    """Nonce value passed to the @clerk/clerk-js script tag for CSP implementation."""

    publishable_key: str = ""
    """The Clerk Publishable Key for your instance, found on the API keys page in the Clerk Dashboard."""

    # proxy_url: str | JSCallable[[str], str] = ""
    proxy_url: str = ""
    """The URL of the proxy server to use for Clerk API requests."""

    router_push: JSCallable[[str], None | Any] | None = None
    """A function to push a new route into the history stack for navigation."""

    router_replace: JSCallable[[str], None | Any] | None = None
    """A function to replace the current route in the history stack for navigation."""

    # sdk_metadata: dict[str, str] = {"name": "", "version": "", "environment": ""}
    # """Contains information about the SDK that the host application is using."""

    # select_initial_session: Callable[[Any], None | Any] | None = None
    # """Function to override the default behavior of using the last active session during client initialization."""

    sign_in_fallback_redirect_url: str = "/"
    """The fallback URL to redirect to after the user signs in if there's no redirect_url in the path."""

    sign_up_fallback_redirect_url: str = "/"
    """The fallback URL to redirect to after the user signs up if there's no redirect_url in the path."""

    sign_in_force_redirect_url: str = ""
    """URL to always redirect to after the user signs in."""

    sign_up_force_redirect_url: str = ""
    """URL to always redirect to after the user signs up."""

    sign_in_url: str = ""
    """URL used for any redirects that might happen, pointing to your primary application on the client-side."""

    sign_up_url: str = ""
    """URL used for any redirects that might happen, pointing to your primary application on the client-side."""

    standard_browser: bool = True
    """Indicates whether ClerkJS assumes cookies can be set (browser setup)."""

    support_email: str = ""
    """Optional support email for display in authentication screens."""

    sync_host: str = ""
    """URL of the web application that the Chrome Extension will sync the authentication state from."""

    telemetry: bool | dict[str, bool] | None = None
    """Controls whether Clerk will collect telemetry data."""

    touch_session: bool = True
    """Indicates whether the Clerk Frontend API touch endpoint is called during page focus to keep the last active session alive."""

    waitlist_url: str = ""
    """The full URL or path to the waitlist page."""

    @classmethod
    def create(cls, *children, **props) -> "ClerkProvider":
        return super().create(*children, **props)

    def add_custom_code(self) -> list[str]:
        return []


def on_load(on_load_events: EventType[()] | None) -> list[IndividualEventType[()]]:
    """Use this to wrap any on_load events that should happen after Clerk has checked authentication.

    Args:
        on_load_events: The events to run after authentication is checked.

    Examples:
        app.add_page(..., on_load=clerk.on_load(<events>))
    """
    if on_load_events is None:
        return []
    on_load_list = (
        on_load_events if isinstance(on_load_events, list) else [on_load_events]
    )

    # Add the on_load events to a registry in the ClerkState instead of actually passing them to on_load.
    #  Then, the wait_for_auth_check event will return the on_load events once auth_checked is True.
    #  Can't just use a blocking wait_for_auth_check because we are really waiting for the frontend event trigger to run,
    #  so we need to not block that while we wait.
    uid = uuid.uuid4()
    ClerkState._set_on_load_events(uid, on_load_list)
    return [ClerkState.wait_for_auth_check(uid)]


T = TypeVar("T", bound=rx.State)


async def _get_state_within_handler(
    current_state: rx.State, desired_state: type[T]
) -> T:
    """Get the desired state from within an event handler.

    Note: Need to be in `async with self` block if called from background event.
    """
    try:
        state = await current_state.get_state(desired_state)
    except ImmutableStateError:
        async with current_state:
            state = await current_state.get_state(desired_state)
    return state


async def get_user(current_state: rx.State) -> clerk_backend_api.models.User:
    """Get the User object from Clerk given the currently logged in user.

    Note: Must be used within an event handler in order to get the appropriate clerk_state.

    Args:
        current_state: The `self` state from the current event handler.

    Examples:

    ```python
    class State(rx.State):
        @rx.event
        async def handle_getting_user_email(self) -> EventType:
            user = await clerk.get_user(self)
            return rx.toast.info(f"User: {user.email}")
    ```
    """
    clerk_state = await _get_state_within_handler(current_state, ClerkState)
    user_id = clerk_state.user_id
    if user_id is None:
        raise MissingUserError("No user_id to get user for")
    user = await clerk_state.client.users.get_async(user_id=user_id)
    if user is None:
        raise MissingUserError("No user found")
    return user


def register_on_auth_change_handler(handler: EventCallback) -> None:
    """Register a handler to be called any time the user logs in or out.

    Args:
        handler: The event handler function to be called.
    """
    ClerkState.register_dependent_handler(handler)


def clerk_provider(
    *children,
    publishable_key: str,
    secret_key: str | None = None,
    register_user_state: bool = False,
    appearance: Appearance | None = None,
    **props,
) -> rx.Component:
    """
    Create a ClerkProvider component to wrap your app/page that uses clerk authentication.

    Note: can also use `wrap_app` to wrap the entire app.

    Args:
        children: The children components to wrap.
        publishable_key: The Clerk Publishable Key for your instance.
        secret_key: Your Clerk app's Secret Key, which you can find in the Clerk Dashboard. It will be prefixed with sk_test_ in development instances and sk_live_ in production instances. Do not expose this on the frontend with a public environment variable.
        register_user_state: Whether to register the ClerkUser state to automatically load user information on login.
        appearance: Optional object to style your components. Will only affect Clerk components.
    """
    if secret_key:
        ClerkState._set_secret_key(secret_key)

    if register_user_state:
        register_on_auth_change_handler(ClerkUser.load_user)

    return ClerkProvider.create(
        ClerkSessionSynchronizer.create(*children),
        publishable_key=publishable_key,
        appearance=appearance,
        **props,
    )


def wrap_app(
    app: rx.App,
    publishable_key: str,
    secret_key: str | None = None,
    register_user_state: bool = False,
    appearance: Appearance | None = None,
    **props,
) -> rx.App:
    """Wraps the entire app with the ClerkProvider.

    For multi-page apps where all pages require Clerk authentication components (including knowing if the user
    is **not** signed in).

    Args:
        app: The Reflex app to wrap.
        publishable_key: The Clerk Publishable Key for your instance.
        secret_key: Your Clerk app's Secret Key. (not needed for frontend only)
        register_user_state: Whether to register the ClerkUser state to automatically load user information on login.
    """
    # 1 makes this the first wrapper around the content
    #  (0 would place it after, 100 would also wrap default reflex wrappers)
    app.app_wraps[(1, "ClerkProvider")] = lambda _: clerk_provider(
        publishable_key=publishable_key,
        secret_key=secret_key,
        register_user_state=register_user_state,
        appearance=appearance,
        **props,
    )
    return app



from .authentication_components import sign_in, sign_up
from .clerk_provider import (
    ClerkState,
    ClerkUser,
    clerk_provider,
    on_load,
    register_on_auth_change_handler,
    wrap_app,
)
from .control_components import (
    clerk_loaded,
    clerk_loading,
    protect,
    redirect_to_user_profile,
    signed_in,
    signed_out,
)
from .pages import add_sign_in_page, add_sign_up_page
from .unstyled_components import (
    SignInButton,
    sign_in_button,
    sign_out_button,
    sign_up_button,
)
from .user_components import user_button, user_profile

__all__ = [
    "ClerkState",
    "ClerkUser",
    "SignInButton",
    "add_sign_in_page",
    "add_sign_up_page",
    "clerk_loaded",
    "clerk_loading",
    "clerk_provider",
    "on_load",
    "protect",
    "redirect_to_user_profile",
    "register_on_auth_change_handler",
    "sign_in",
    "sign_in_button",
    "sign_out_button",
    "sign_up",
    "sign_up_button",
    "signed_in",
    "signed_out",
    "user_button",
    "user_profile",
    "wrap_app",
]



from typing import Any, Literal

from reflex.components.props import PropsBase

LiteralBaseTheme = Literal["default", "dark", "neobrutalism", "shadesOfPurple"]


class Layout(PropsBase):
    animations: bool = True
    """Whether to enable animations inside the components. Defaults to true."""

    help_page_url: str = ""
    """The URL to your help page."""

    logo_image_url: str = ""
    """The URL to your logo image."""

    logo_link_url: str = ""
    """Controls where the browser will redirect to after the user clicks the application logo."""

    logo_placement: str = "inside"
    """The placement of your logo. Defaults to 'inside'."""

    privacy_page_url: str = ""
    """The URL to your privacy page."""

    shimmer: bool = True
    """Enables the shimmer animation for avatars. Defaults to true."""

    show_optional_fields: bool = True
    """Whether to show optional fields on the sign in and sign up forms. Defaults to true."""

    social_buttons_placement: str = "top"
    """The placement of your social buttons. Defaults to 'top'."""

    social_buttons_variant: str = "auto"
    """The variant of your social buttons."""

    terms_page_url: str = ""
    """The URL to your terms page."""

    unsafe_disable_development_mode_warnings: bool = False
    """Whether development warnings show up in development mode."""


class Variables(PropsBase):
    color_primary: str = ""
    color_danger: str = ""
    color_success: str = ""
    color_warning: str = ""
    color_neutral: str = ""
    color_text: str = ""
    color_text_on_primary_background: str = ""
    color_text_secondary: str = ""
    color_background: str = ""
    color_input_text: str = ""
    color_input_background: str = ""
    color_shimmer: str = ""
    font_family: str = "inherit"
    font_family_buttons: str = "inherit"
    font_size: str = "0.8125rem"
    font_weight: dict[str, int] = {
        "normal": 400,
        "medium": 500,
        "semibold": 600,
        "bold": 700,
    }
    border_radius: str = "0.375rem"
    spacing_unit: str = "1rem"


class Captcha(PropsBase):
    theme: str = "auto"
    """The CAPTCHA widget theme. Defaults to auto."""

    size: str = "normal"
    """The CAPTCHA widget size. Defaults to normal."""

    language: str = ""
    """The CAPTCHA widget language/locale."""


class Appearance(PropsBase):
    # TODO: This needs to reference the actual theme object, not a string -- don't know how to do that yet.
    # base_theme: LiteralBaseTheme = "default"
    # """A theme used as the base theme for the components."""

    layout: Layout | None = None
    """Configuration options that affect the layout of the components."""

    variables: Variables | None = None
    """General theme overrides."""

    elements: dict[str, Any] | None = None
    """Fine-grained theme overrides."""

    captcha: Captcha | None = None
    """Configuration options that affect the appearance of the CAPTCHA widget."""
